/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2016                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
#pragma once

#include <deque>
#include <functional>
#include <random>
#include <vcg/simplex/face/pos.h>
#include <vcg/simplex/face/topology.h>
#include <vcg/complex/algorithms/update/quality.h>

namespace vcg{
namespace tri{

template <class MeshType>
class RansacPlaneFitting{
  public:
  typedef typename MeshType::VertexType VertexType;
  typedef typename MeshType::ScalarType  ScalarType;
  typedef typename MeshType::FacePointer FacePointer;
class Stats
  {
public:
      int improvedCnt;
    int correctGuessCnt;
      
      void reset()
    {
          improvedCnt=0;
        correctGuessCnt=0;
      }
};
      
class Parameters
{
  public:
  
  float  planeDistanceThresholdPerc = 0.1f; // distance threshold for inlier/outlier classification
  float  planeDistanceThresholdAbs = -1.0f; // distance threshold for inlier/outlier classification (neg values means not initialized) 
  float sampleDistanceThresholdPerc = 0.1f; // distance used to be sure that sampling points are far apart
  float sampleDistanceThresholdAbs = -1.0f; // distance used to be sure that sampling points are far apart
  int  maxIterations = 1000; // the number of planes we sample to search the best one. The higher the number the more accurate guess we get.
  float  minInlierFraction = 0.5f; // minimum fraction of inliers to accept a plane
  int quickInlierEstimationSamples = 500; // number of samples used for the quick estimate of the inlier fraction
  bool onlySelected = false; // if true, only selected vertices are used for the plane fitting
  Stats s;
  Parameters() {}

  std::vector<int> permutationVec;
  void Init(MeshType &mesh)
  {
    // let fill the permuation vector with the indices of the selected vertices

    for(int i=0;i<mesh.VN();i++)
        if(!onlySelected || mesh.vert[i].IsS())
              permutationVec.push_back(i);

    std::random_device rd;
    std::mt19937 g(rd());
    std::shuffle(permutationVec.begin(),permutationVec.end(),g);
    
    if(planeDistanceThresholdAbs < 0)
        planeDistanceThresholdAbs=   mesh.bbox.Diag()*  planeDistanceThresholdPerc;
    if(sampleDistanceThresholdAbs < 0)
        sampleDistanceThresholdAbs= mesh.bbox.Diag()*sampleDistanceThresholdPerc;
  }
};

static math::MarsenneTwisterRNG &rnd() 
{
  static math::MarsenneTwisterRNG rnd;
  return rnd;
}

static void SearchBestFittingPlane(MeshType &mesh, Parameters &param, Plane3f &bestFittingPlane)
{
  vcg::Plane3f fitPlane;
  float inliersFraction = 0;
  int ransacIterations = 0;
  float bestInliersFraction = 0;
  param.s.reset();
  
  while (ransacIterations < param.maxIterations)
  {
    GeneratePlane(mesh, param, fitPlane);
    EstimateInlierFraction(mesh, param, fitPlane,  inliersFraction);
    
    if (inliersFraction > bestInliersFraction * 0.8f)
    {
        ComputeInlierFraction(mesh, param, fitPlane, inliersFraction);  // this is the precise estimate of the inlier fraction
        ++param.s.correctGuessCnt;
    }
    
    if (inliersFraction > bestInliersFraction)
    {
        ++param.s.improvedCnt;
      bestInliersFraction = inliersFraction;
      bestFittingPlane = fitPlane;
    }
    ransacIterations++;
  }
}

static void GeneratePlane(MeshType &mesh, Parameters &param, Plane3f &fittingPlane)
{
  // First I take three random points over my mesh
  // checking that they are far apart and not aligned
 
  int p0,p1,p2;
  p0 = param.permutationVec[rnd().generate(param.permutationVec.size())];
  while(true)
  {
    p1 = param.permutationVec[rnd().generate(param.permutationVec.size())];
    if (Distance(mesh.vert[p0].P(),mesh.vert[p1].P()) > param.sampleDistanceThresholdAbs)  
      break;
  }
 while(true)
  {
    p2 = param.permutationVec[rnd().generate(param.permutationVec.size())];
    
    if ( (Distance(mesh.vert[p0].P(),mesh.vert[p2].P()) > param.sampleDistanceThresholdAbs) &&
    (Distance(mesh.vert[p1].P(),mesh.vert[p2].P()) > param.sampleDistanceThresholdAbs) )
      break;
  }
 // Then I compute the plane passing through them
  fittingPlane.Init(mesh.vert[p0].P(),mesh.vert[p1].P(),mesh.vert[p2].P());
}

static void ComputeInlierFraction(MeshType &mesh, Parameters &param, const Plane3f &fittingPlane, float &inlierFraction)
{
  EstimateInlierFraction(mesh, param, fittingPlane, inlierFraction, true);
}

// This function computes an estimate of the inlier fraction of a plane by choosing a random subset of points to check
static void EstimateInlierFraction(MeshType &mesh, Parameters &param, const Plane3f &fittingPlane, float &inlierFraction, bool useAllPoints=false)
{
 int pointsOnPlane = 0;
  int checkingPoints = param.permutationVec.size();
 if(!useAllPoints)
    checkingPoints = std::min(checkingPoints,param.quickInlierEstimationSamples);
 
  for(int i=0;i<checkingPoints;++i)
  {
    int p = param.permutationVec[i];
    if (fabs(SignedDistancePlanePoint(fittingPlane,mesh.vert[p].P())) < param.planeDistanceThresholdAbs)
        pointsOnPlane++;
  }
  inlierFraction = (float)pointsOnPlane/(float)checkingPoints; 
}

static void DeselectMeshByPlaneDistance(MeshType &mesh, Parameters &param, Plane3f &fittingPlane)
{
  for (int i = 0; i < mesh.vert.size(); i++)
  {
    if (fabs(SignedDistancePlanePoint(fittingPlane,mesh.vert[i].P())) < param.planeDistanceThresholdAbs)
      mesh.vert[i].ClearS();
  }
}

static void ColorizeMeshByPlaneDistance(MeshType &mesh, Parameters &param, Plane3f &fittingPlane, Color4b c=Color4b::Red)
{
  for (int i = 0; i < mesh.vert.size(); i++)
  {
    if (fabs(SignedDistancePlanePoint(fittingPlane,mesh.vert[i].P())) < param.planeDistanceThresholdAbs)
      mesh.vert[i].C() = c;
  }
}



};
}   // end namespace tri
}   // end namespace vcg
